<?php
/**
 * @file
 * Connector to Memsource Service.
 */

/**
 * Memsource connector class.
 */
class MemsourceConnector {

  private $translator;

  private $serviceUrl;

  private $userName;

  private $password;

  /** @var string|null auth api token */
  private $token;

  /**
   * @param string $service_url
   *   Memsource API URL.
   * @param $user_name
   *   Memsource Cloud user name.
   * @param $password
   *   Memsource Cloud password.
   * @param $translator
   */
  function __construct($service_url, $user_name, $password, $translator) {
    $this->serviceUrl = $service_url;
    $this->userName = $user_name;
    $this->password = $password;
    $this->translator = $translator;
  }

  /**
   * {@inheritdoc}
   */
  public function getSupportedRemoteLanguages() {
    $supported_remote_languages = [];

    if (empty($this->userName) || empty($this->password)) {
      return $supported_remote_languages;
    }

    try {
      $supported_languages = $this->sendApiRequest('v2/language/listSupportedLangs');
      foreach ($supported_languages as $language) {
        $supported_remote_languages[$language['code']] = $language['name'];
      }
    } catch (\Exception $e) {
      // Ignore exception, nothing we can do.
    }
    asort($supported_remote_languages);
    return $supported_remote_languages;
  }

  /**
   * {@inheritdoc}
   */
  public function requestJobItemsTranslation(array $job_items) {
    /** @var TMGMTJob $job */
    $job = reset($job_items)->getJob();
    $project_id = 0;
    $due_date_array = $job->getSetting('due_date');
    $date_time = new DateTime();
    $date_time->setDate($due_date_array['year'], $due_date_array['month'], $due_date_array['day']);
    $due_date = $date_time->format('Y-m-d H:i');
    try {
      $project_id = $this->newTranslationProject($job, $due_date);
      $job->addMessage('Created a new project in Memsource Cloud with the id: @id', ['@id' => $project_id], 'debug');

      /** @var TMGMTJobItem $job_item */
      foreach ($job_items as $job_item) {
        $file_id = $this->sendFiles($job_item, $project_id, $due_date);
        /** @var TMGMTRemote $remote_mapping */
        $remote_mapping = entity_create('tmgmt_remote', [
          'tjid' => $job->tjid,
          'tjiid' => $job_item->tjiid,
          'remote_identifier_1' => 'tmgmt_memsource',
          'remote_identifier_2' => $project_id,
          'remote_identifier_3' => $file_id,
          'remote_data' => [
            'FileStateVersion' => 1,
            'TmsState' => 'TranslatableSource',
            'RequiredBy' => $due_date,
          ],
        ]);
        $remote_mapping->save();
      }
    } catch (TMGMTException $e) {
      try {
        $this->sendFileError('RestartPoint03', $project_id, '', $job, $due_date, $e->getMessage(), TRUE);
      } catch (TMGMTException $e) {
        watchdog('tmgmt_memsource', 'Error sending the error file: @error', ['@error' => $e->getMessage()], $severity = WATCHDOG_ERROR);
      }
      $job->rejected('Job has been rejected with following error: @error',
        ['@error' => $e->getMessage()], 'error');
      if (isset($remote_mapping)) {
        $remote_mapping->delete();
      }
    }
    return $job;
  }

  /**
   * Performs a login to Memsource Cloud.
   *
   * @return bool
   *   A success or failure.
   */
  public function loginToMemsource() {
    $params = [
      'username' => $this->userName,
      'password' => $this->password,
    ];
    try {
      $result = $this->request('v2/auth/login', 'GET', $params);
      if ($result['token']) {
        // Store the token.
        $this->storeToken($result['token']);
        return TRUE;
      }
    } catch (TMGMTException $ex) {
      if (variable_get('tmgmt_memsource_debug', FALSE)) {
        watchdog('tmgmt_memsource', 'Unable to log in to Memsource API: ' . $ex->getMessage(), WATCHDOG_DEBUG);
      }
    }
    return FALSE;
  }

  /**
   * Stores a Memsource API token.
   *
   * @param string $token
   *   Token.
   */
  public function storeToken($token) {
    $this->token = $token;
  }

  /**
   * Returns a Memsource API token.
   *
   * @return string
   *   Token.
   */
  public function getToken() {
    return $this->token;
  }

  /**
   * Verifies a Memsource API token validity.
   */
  public function verifyToken() {
    if (!$this->translator) {
      throw new TMGMTException('There is no Translator entity. Access to the Memsource API is not possible.');
    }
    $code = $this->request('v3/auth/whoAmI', 'GET', ['token' => $this->getToken()], FALSE, TRUE);
    if ($code != 200) {
      // Token is invalid, try to re-login.
      $this->loginToMemsource();
    }
  }

  /**
   * Sends a request to the Memsource API and refreshes the token if necessary.
   *
   * @param string $path
   *   API path.
   * @param string $method
   *   (Optional) HTTP method.
   * @param array $params
   *   (Optional) API params.
   * @param bool $code
   *   (Optional) If true, return only the response HTTP status code.
   * @param bool $download
   *   (Optional) If true, return the response body as a downloaded content.
   * @param string $body
   *   (Optional) An optional request body.
   *
   * @return array|int|null
   *   Result of the API request.
   *
   * @throws TMGMTException
   */
  public function sendApiRequest($path, $method = 'GET', $params = [], $code = FALSE, $download = FALSE, $body = NULL) {
    $result = NULL;
    $params['token'] = $this->getToken();
    try {
      $result = $this->request($path, $method, $params, $code, $download, $body);
    } catch (TMGMTException $ex) {
      if ($ex->getCode() == 401) {
        // Token is invalid, try to re-login.
        $this->loginToMemsource();
        $params['token'] = $this->getToken();
        $result = $this->request($path, $method, $params, $code, $download, $body);
      }
      else {
        throw $ex;
      }
    }
    return $result;
  }

  /**
   * Does a request to memsource services.
   *
   * @param string $path
   *   Resource path.
   * @param string $method
   *   (Optional) HTTP method (GET, POST...). By default uses GET method.
   * @param array $params
   *   (Optional) Form parameters to send to memsource service.
   * @param bool $code
   *   (Optional) If we want to return the status code of the call. FALSE by
   *   default.
   * @param bool $download
   *   (Optional) If we expect resource to be downloaded. FALSE by default.
   * @param string $body
   *   (Optional) Body of the POST request. NULL by
   *   default.
   *
   * @return array|int|null
   *   Response array from memsource.
   *
   * @throws TMGMTException
   */
  public function request($path, $method = 'GET', $params = [], $code = FALSE, $download = FALSE, $body = NULL) {
    $options = [];

    $url = $this->serviceUrl . '/' . $path;

    if ($body) {
      $options['data'] = $body;
    }
    else {
      if ($method == 'GET') {
        $url .= '?' . http_build_query($params);
        //        $options['query'] = $params;
      }
      else {
        $options['headers']['Content-Type'] = 'application/json';
        $options['data'] = json_encode($params);
        $options['method'] = $method;
      }
    }

    $response = drupal_http_request($url, $options);
    $received_data = $response->data;

    if ($response->code != 200) {
      watchdog('tmgmt_memsource', '%method Request to %url:<br>
        <ul>
            <li>Request: %request</li>
            <li>Response: %response</li>
        </ul>
        ', [
        '%method' => $method,
        '%url' => $url,
        '%request' => json_encode($options),
        '%response' => $received_data,
      ], WATCHDOG_ERROR);
      if ($code) {
        return $response->code;
      }
      throw new TMGMTException('Unable to connect to Memsource service due to following error: @error at @url',
        ['@error' => $response->code, '@url' => $url], $response->code);
    }

    watchdog('tmgmt_memsource', '%method Request to %url:<br>
        <ul>
            <li>Request: %request</li>
            <li>Response: %response</li>
        </ul>
        ', [
      '%method' => $method,
      '%url' => $url,
      '%request' => json_encode($options),
      '%response' => $received_data,
    ], WATCHDOG_DEBUG);

    if ($code) {
      return $response->code;
    }

    // If we are expecting a download, just return received data.
    if ($download) {
      return $received_data;
    }
    $received_data = json_decode($received_data, TRUE);

    return $received_data;
  }

  /**
   * Creates new translation project at Memsource Cloud.
   *
   * @param TMGMTJob $job
   *   The job.
   * @param string $due_date
   *   The date by when the translation is required.
   *
   * @return int
   *   Memsource project id.
   *
   * @throws TMGMTException
   */
  public function newTranslationProject(TMGMTJob $job, $due_date) {
    $name = $job->getSetting('project_name') ?: 'Drupal TMGMT project ' . $job->tjid;
    $params = [
      'name' => $name,
      'sourceLang' => $job->source_language,
      'targetLang' => $job->target_language,
      'dateDue' => $due_date,
    ];
    $template_id = $job->getSetting('project_template');
    if ($template_id == '0') {
      $result = $this->sendApiRequest('v4/project/create', 'GET', $params);
    }
    else {
      $params['template'] = $template_id;
      $result = $this->sendApiRequest('v4/project/createFromTemplate', 'GET', $params);
    }
    return $result['id'];
  }

  /**
   * Send the files to Memsource Cloud.
   *
   * @param TMGMTJobItem $job_item
   *   The Job.
   * @param int $project_id
   *   Thebigword project id.
   * @param string $due_date
   *   The date by when the translation is required.
   *
   * @return int
   *   Memsource jobPartId.
   */
  public function sendFiles(TMGMTJobItem $job_item, $project_id, $due_date) {
    // Prepare parameters for Job API.
    $xliff_converter = tmgmt_file_format_controller('xlf');

    $job_item_id = $job_item->tjiid;
    $target_language = $job_item->getJob()->target_language;
    $conditions = ['tjiid' => ['value' => $job_item_id]];
    $xliff = $xliff_converter->export($job_item->getJob(), $conditions);
    $name = "JobID_{$job_item->tjid}_JobItemID_{$job_item_id}_{$job_item->getJob()->source_language}_{$target_language}";

    $file_id = $this->uploadFileResource($xliff, $name);

    $job_part_id = $this->createJob($project_id, $file_id, $target_language, $due_date);

    return $job_part_id;
  }

  /**
   * Creates a file resource at Memsource Cloud.
   *
   * @param string $xliff
   *   .XLIFF string to be translated. It is send as a file.
   * @param string $name
   *   File name of the .XLIFF file.
   *
   * @return string
   *   Memsource uuid of the resource.
   *
   * @throws TMGMTException
   */
  public function uploadFileResource($xliff, $name) {
    $file_name = $name . '.xliff';
    $form_params = [
      'token' => $this->getToken(),
      'name' => $file_name,
    ];
    $result = $this->sendApiRequest('v2/file/uploadRequestBody?' . http_build_query($form_params), 'POST', [], FALSE, FALSE, $xliff);
    return $result['uid'];
  }

  /**
   * Creates a new job at Memsource Cloud.
   *
   * @param string $project_id
   *   Project ID.
   * @param string $file_id
   *   File ID.
   * @param string $target_language
   *   Target language code.
   * @param string $due_date
   *   Job due date.
   *
   * @return string
   *   Memsource uuid of the resource.
   *
   * @throws TMGMTException
   */
  public function createJob($project_id, $file_id, $target_language, $due_date) {
    $form_params = [
      'project' => $project_id,
      'file' => $file_id,
      'targetLang' => $target_language,
      'due' => $due_date,
      'fileFormat' => 'wpxliff',
    ];
    $result = $this->sendApiRequest('v8/job/create', 'GET', $form_params);
    return $result['jobParts'][0]['id'];
  }

  /**
   * Parses received translation from Memsource and returns unflatted data.
   *
   * @param string $data
   *   Base64 encode data, received from Memsource.
   *
   * @return array
   *   Unflatted data.
   */
  protected function parseTranslationData($data) {
    $xlf_converter = tmgmt_file_format_controller('xlf');
    $xlf_converter = new $xlf_converter;
    // Import given data using XLIFF converter. Specify that passed content is
    // not a file.
    return $xlf_converter->import($data, FALSE);
  }

  /**
   * Fetches translations for job items of a given job.
   *
   * @param TMGMTJob $job
   *   A job containing job items that translations will be fetched for.
   *
   * @return array
   *   Array containing a containing the number of items translated and the
   *   number that has not been translated yet.
   */
  public function fetchTranslatedFiles(TMGMTJob $job) {
    $translated = 0;
    $errors = [];

    try {
      /** @var TMGMTJobItem $job_item */
      foreach ($job->getItems() as $job_item) {
        /** @var TMGMTRemoteController $remote_mapping_controller */
        $remote_mapping_controller = entity_get_controller('tmgmt_remote');
        $mappings = $remote_mapping_controller->loadByLocalData($job_item->tjid, $job_item->tjiid);
        /** @var TMGMTRemote $mapping */
        foreach ($mappings as $mapping) {
          $job_part_id = $mapping->remote_identifier_3;
          $project_id = $mapping->remote_identifier_2;
          $old_state = $mapping->getRemoteData('TmsState');
          $params = [
            'jobPart' => $job_part_id,
          ];
          $info = [];
          try {
            $info = $this->sendApiRequest('v8/job/get', 'GET', $params);
          } catch (TMGMTException $e) {
            $job->addMessage('Error fetching the job item: @job_item. Memsource job @job_part_id not found.',
              [
                '@job_item' => $job_item->label(),
                '@job_part_id' => $job_part_id,
              ], 'error');
            $errors[] = 'Memsource job ' . $job_part_id . ' not found, it was probably deleted.';
          }
          if (array_key_exists('status', $info)) {
            if ($this->remoteTranslationCompleted($info['status'])) {
              try {
                $this->addFileDataToJob($job, $info['status'], $project_id, $job_part_id);
                $translated++;
              } catch (TMGMTException $e) {
                $restart_point = $old_state == 'TranslatableSource' ? 'RestartPoint01' : 'RestartPoint02';
                $this->sendFileError($restart_point, $project_id, $job_part_id, $job_item->getJob(), $mapping->getRemoteData('RequiredBy'), $e->getMessage());
                $job->addMessage('Error fetching the job item: @job_item.', ['@job_item' => $job_item->label()], 'error');
                continue;
              }
            }
          }
        }
      }
    } catch (TMGMTException $e) {
      watchdog('tmgmt_memsource', 'Could not pull translation resources: @error', ['@error' => $e->getMessage()], $severity = WATCHDOG_ERROR);
    }
    return [
      'translated' => $translated,
      'untranslated' => count($job->getItems()) - $translated,
      'errors' => $errors,
    ];
  }

  /**
   * Retrieve all the updates for all the job items in a translator.
   *
   * @param TMGMTJobItem $job_item
   *   The translator.
   *
   * @return int
   *   The number of updated items.
   */
  public function pullRemoteTranslation(TMGMTJobItem $job_item) {
    $job = $job_item->getJob();
    $this->translator = $job->getTranslator();
    /** @var TMGMTRemoteController $remote_mapping_controller */
    $remote_mapping_controller = entity_get_controller('tmgmt_remote');
    $remotes = $remote_mapping_controller->loadByLocalData($job->tjid, $job_item->tjiid);
    /** @var TMGMTRemote $remote */
    $remote = reset($remotes);
    $params = [
      'jobPart' => $remote->remote_identifier_3,
    ];
    $info = $this->sendApiRequest('v8/job/get', 'GET', $params);
    if ($this->remoteTranslationCompleted($info['status'])) {
      try {
        $this->addFileDataToJob($job, $info['status'], $remote->remote_identifier_2, $remote->remote_identifier_3);
        return 1;
      } catch (TMGMTException $e) {
        $restart_point = 'RestartPoint02';
        $this->sendFileError($restart_point, $remote->remote_identifier_2, $remote->remote_identifier_3, $job, $remote->getRemoteData('RequiredBy'), $e->getMessage());
        $job->addMessage('Error fetching the job item: @job_item.', [
          '@job_item' => $remote->getJobItem()
            ->label(),
        ], 'error');
      }
    }
    return 0;
  }

  /**
   * Checks if the translation has one of the completed statuses.
   *
   * @param string $status
   *   Status code.
   *
   * @return bool
   *   True if completed.
   */
  public function remoteTranslationCompleted($status) {
    return $status == 'COMPLETED_BY_LINGUIST' || $status == 'COMPLETED';
  }

  /**
   * Sends an error file to memsource.
   *
   * @param string $state
   *   The state.
   * @param int $project_id
   *   The project id.
   * @param string $file_id
   *   The file id to update.
   * @param TMGMTJob $job
   *   The Job.
   * @param string $required_by
   *   The date by when the translation is required.
   * @param string $message
   *   (Optional) The error message.
   * @param bool $confirm
   *   (Optional) Set to TRUE if also want to send the confirmation message
   *   of this error. Otherwise will not send it.
   *
   * @throws TMGMTException
   *   If there is a problem with the request.
   */
  public function sendFileError($state, $project_id, $file_id, TMGMTJob $job, $required_by, $message = '', $confirm = FALSE) {
    // Use this function to handle the error at Memsource side (not used now).
  }

  /**
   * Retrieve the data of a file in a state.
   *
   * @param TMGMTJob $job
   *   The Job to which will be added the data.
   * @param string $state
   *   The state of the file.
   * @param int $project_id
   *   The project id.
   * @param string $file_id
   *   The file ID.
   *
   * @throws TMGMTException
   */
  public function addFileDataToJob(TMGMTJob $job, $state, $project_id, $file_id) {
    $params = [
      'jobPart' => $file_id,
    ];
    $data = $this->sendApiRequest('v8/job/getCompletedFile', 'GET', $params, FALSE, TRUE);
    $decoded_data = $data;
    $file_data = $this->parseTranslationData($decoded_data);
    $job->addTranslatedData($file_data, []);
    /** @var TMGMTRemoteController $remote_mapping_controller */
    $remote_mapping_controller = entity_get_controller('tmgmt_remote');
    $mappings = $remote_mapping_controller->loadByRemoteIdentifier('tmgmt_memsource', $project_id, $file_id);
    /** @var TMGMTRemote $mapping */
    $mapping = reset($mappings);
    $mapping->removeRemoteData('TmsState');
    $mapping->addRemoteData('TmsState', $state);
    $mapping->save();
  }

}
